/*=============================================================================
segvtracer.c - Tracing and recording segfaults
  Copyright (C) Jan "Hirogen2" Engelhardt <hirogen2 at gmx de>, 2003
  -- License restrictions apply (Dual FAL/GPL2).
  -- For details see doc/FAL.txt and doc/GPL2.txt.
===============================================================================
Oh, and this works only for x86, x86_64 (AMD64).
Thanks to "The Art Of Assembly" for the general knowledge,
thanks to SuSE, AMD and SourceForge for providing an AMD64 host.
-----------------------------------------------------------------------------*/
#define _GNU_SOURCE 1
#include <sys/ptrace.h>
#include <sys/select.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <errno.h>
#include <fcntl.h>
#include <pthread.h>
#include <signal.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <linux/user.h> /* must come after <sys/ptrace.h> */
#include <libHX.h>

#define DEV_CURTTY "/dev/tty"
#define DEV_PTMX "/dev/ptmx"
#define LOGFILE "segvtracer.log"

// EFLAGS register bitfields
#define FLAG_OVF(v, c)  (((v) & (1 << 11)) ? (c) : '.') /* overflow */
#define FLAG_DIR(v, c)  (((v) & (1 << 10)) ? (c) : '.') /* direction */
#define FLAG_INTR(v, c)  (((v) & (1 << 9)) ? (c) : '.') /* interrupt */
#define FLAG_SIGN(v, c)  (((v) & (1 << 7)) ? (c) : '.') /* sign */
#define FLAG_ZERO(v, c)  (((v) & (1 << 6)) ? (c) : '.') /* zero */
#define FLAG_AUXC(v, c)  (((v) & (1 << 4)) ? (c) : '.') /* auxilary carry */
#define FLAG_PAR(v, c)   (((v) & (1 << 2)) ? (c) : '.') /* parity */
#define FLAG_CARRY(v, c) (((v) & (1 << 0)) ? (c) : '.') /* carry flag */

#define MAXFNLEN 1024 /* max length for filename buffer */
#define MAXLNLEN 1024 /* max length for line buffer */

typedef unsigned char byte;

#ifdef PURE /* "command line mode" */
#    define _segvtracer_main_pf int main
#else
     // only run constructor if in "hook mode"
     static void _segvtracer_hook(void) __attribute__((constructor));
#    define _segvtracer_main_pf static int _segvtracer_main
#endif

_segvtracer_main_pf(int, char **); // foo proto

static void block_signals(void);
static int child_branch(int *, char **);
static int handle_exit(pid_t, int);
static int handle_killsig(pid_t, int);
static int handle_signal(pid_t, int);
static void dump_eip(pid_t);
static void dump_maps(pid_t);
static void dump_regs(pid_t);
static void dump_stack(pid_t);
static int logmsg(char *, ...);
static void *xlogger(void *);

static FILE *logp = NULL, *conp = NULL;

//-----------------------------------------------------------------------------
#ifndef PURE
static void _segvtracer_hook(void) {
    int running_under_something = 0, i;
    for(i = 3; i < 256; ++i) {
        if(!(lseek(i, 0, SEEK_CUR) < 0 && errno == EBADF)) {
            ++running_under_something;
            break;
        }
    }

    if(!running_under_something) {
        // no GDB running (dirty hack tho)
        char *argv[] = {"-", NULL};
        int rv = _segvtracer_main(1, argv);
        switch(rv) {
            case 2: // continue
                break;
            default:
                // rv = 0: painful exit, rv = 1: normal exit
                exit(!rv);
        }
    }
    return;
}
#endif

_segvtracer_main_pf(int argc, char **argv) {
    pthread_t thr_xlog;
    int i, ptm[2];
    pid_t pid;

    ++argv;
    --argc;

    /* Opening a FILE stream for logging, on current console, to not use
    stdout, since that is used by xlogger() and which one might redirect at the
    shell level. */
    if((conp = fopen(DEV_CURTTY, "a+b")) == NULL) {
        fprintf(stderr, "Could not open %s: %s\n", DEV_CURTTY, strerror(errno));
        return 0;
    }

    fprintf(conp, "\e[32mSEGV\e[1m>\e[0m Initialized. " // ]]]
     "(Segvtracer by Jan Engelhardt)\n");

    if((logp = fopen(LOGFILE, "a")) == NULL) {
        logmsg("Could not open %s: %s\n", LOGFILE, strerror(errno));
        return 0;
    }

    /* Buffering is tourned off since fgets() only flushes the stdout stream
    for the application that called it. With our pseudo-terminal pipes, we get
    data as soon as fgets() has internall called the fflush() routine, however,
    _we_ (the tracer) would also need to flush it then. Since it is not
    detectable (to my beliefs), we have to turn off buffering for all time. */
    setvbuf(stdout, NULL, _IONBF, 0);
    setvbuf(stderr, NULL, _IONBF, 0);
    setvbuf(conp, NULL, _IONBF, 0);
    setvbuf(logp, NULL, _IONBF, 0);
    fprintf(logp, "----- New session and/or application -----\n");

    for(i = 0; i < 2; ++i) {
        /* Pseudo-terminals are used, so we are able to log the output to logp,
        if the default stdout is redirected elsewhere. */
        if((ptm[i] = open(DEV_PTMX, O_RDWR | O_NOCTTY, 0600)) < 0) {
            logmsg("Could not open %s: %s\n", DEV_PTMX, strerror(errno));
            return 0;
        }

        grantpt(ptm[i]);
        unlockpt(ptm[i]);
    }

    if((pid = fork()) < 0) {
        logmsg("Could not fork(): %s\n", strerror(errno));
        return 0;
    } else if(pid == 0) {
        return child_branch(ptm, argv);
    }

    block_signals();
    pthread_create(&thr_xlog, NULL, xlogger, (void *)ptm);

    while(pid > 0) {
        int st;
        waitpid(pid, &st, 0);
        if((WIFEXITED(st) && handle_exit(pid, st)) ||
         (WIFSIGNALED(st) && handle_killsig(pid, st)) ||
         (WIFSTOPPED(st) && handle_signal(pid, st))) {
            ptrace(PTRACE_KILL, pid, NULL, NULL);
            pthread_join(thr_xlog, NULL);
            break;
        }
    }

    logmsg("Tracing completed, log stored in segvtracer.log.\n");
    fclose(logp);
    fclose(conp);
    return 1;
}

static void block_signals(void) {
    struct sigaction sa;
    sa.sa_handler = SIG_IGN;
    sa.sa_flags = SA_RESTART;

    /* prevent that a CTRL+C kills the tracer/xlogger (by default signal gets
    sent to a process group (which tracer+appl+xlogger are altogether)). */
    sigaction(SIGINT, &sa, NULL);
    sigaction(SIGQUIT, &sa, NULL);
    sigaction(SIGTERM, &sa, NULL);
    return;
}

static int child_branch(int *ptm, char **argv) {
    char *pts[] = {HX_strdup(ptsname(ptm[0])), HX_strdup(ptsname(ptm[1]))};
    int i, rv;

    for(i = 0; i < 2; ++i) {
        close(ptm[i]);
        /* Fiddle around and try replacing stdout/stderr with the
        pseudo-terminal links. */
        close(i + 1);
        rv = open(pts[i], O_WRONLY, 0600);
        if(rv < 0) {
            logmsg("Could not open %s: %s\n", pts[0], strerror(errno));
            return 0;
        } else if(rv != i + 1) {
            logmsg("%s did not got routed to fd %d, but to fd %d\n",
             pts[i], i + 1, rv);
            return 0;
        }
    }

    ptrace(PTRACE_TRACEME, 0, NULL, NULL);
    logmsg("Launching application (%s)\n", *argv);
#ifdef PURE
    rv = execvp(*argv, (char **)argv);
    logmsg("Unable to do so: %s\n", strerror(errno));
    fclose(logp);
    fclose(conp);
    return rv;
#else
    return 2; // will continue with main()
#endif
}

static int handle_exit(pid_t pid, int st) {
    logmsg("Program exited with value %d\n", WEXITSTATUS(st));
    return 1;
}

static int handle_killsig(pid_t pid, int st) {
    int sig = WTERMSIG(st);
    logmsg("Application terminated due to signal %d (%s)\n",
     sig, strsignal(sig));

    switch(sig) {
        case SIGILL:  // 4: Illegal instruction
        case SIGABRT: // 6: Aborted
        case SIGBUS:  // 7: Bus error
        case SIGFPE:  // 8: Floating-point exception
        case SIGKILL: // 9: Killed
        case SIGSEGV: // 11: Segementation fault
        case SIGPIPE: // 13: Broken pipe
        case 16:      // 16: Stack fault
            dump_regs(pid);
            dump_stack(pid);
            break;
    }

    return 1;
}

static int handle_signal(pid_t pid, int st) {
    int sig = WSTOPSIG(st);
    logmsg("Application received signal %d (%s)\n", sig, strsignal(sig));

    switch(sig) {
        case SIGILL:  // 4: Illegal instruction
        case SIGABRT: // 6: Aborted
        case SIGBUS:  // 7: Bus error
        case SIGFPE:  // 8: Floating-point exception
        case SIGKILL: // 9: Killed ... -> only handled in handle_killsig()
        case SIGSEGV: // 11: Segementation fault
        case SIGPIPE: // 13: Broken pipe
        case 16:      // 16: Stack fault
            dump_maps(pid);
            dump_regs(pid);
            dump_stack(pid);
            return 1;
            break;

        case SIGTRAP: // 5: Trace/Breakpoint trap
            ptrace(PTRACE_CONT, pid, NULL, SIGCONT);
            dump_eip(pid);
            break;

        case SIGCHLD: // 17: Child exited
            dump_eip(pid);
            //dump_stack(pid);
            ptrace(PTRACE_CONT, pid, NULL, sig);
            break;

        default:
            /* Deliver signal to application but at least tell us where it
            happened */
            dump_eip(pid);
            ptrace(PTRACE_CONT, pid, NULL, sig);
            break;
    }

    return 0;
}

static void dump_eip(pid_t pid) {
    struct user_regs_struct rg;
    ptrace(PTRACE_GETREGS, pid, NULL, &rg);
#ifdef __x86_64__
    logmsg("\tRegister RIP = 0x%08lX\n", rg.rip);
#else
    logmsg("\tRegister EIP = 0x%08lX\n", rg.eip);
#endif
    return;
}

static void dump_maps(pid_t pid) {
    FILE *fp;
    char buf_a[MAXLNLEN], buf_b[MAXLNLEN], buf_f[MAXFNLEN];
    snprintf(buf_f, MAXFNLEN, "/proc/%d/maps", pid);
    if((fp = fopen(buf_f, "r")) == NULL) {
        logmsg("No map information available\n");
    }

    logmsg("Maps dump =>\n");
    while(fgets(buf_a, MAXLNLEN, fp) != NULL) {
        HX_chomp(buf_a);
        snprintf(buf_b, MAXLNLEN, "\t%s\n", buf_a);
        logmsg(buf_b);
    }

    return;
}

static void dump_regs(pid_t pid) {
    struct user_regs_struct rg;
    ptrace(PTRACE_GETREGS, pid, NULL, &rg);

    logmsg("Register dump =>\n");
#ifdef __x86_64__
    logmsg("\tRAX = 0x%08lX  CS  = 0x%08lX     RIP = 0x%08lX\n", rg.rax, rg.cs, rg.rip);
    logmsg("\tRBX = 0x%08lX  DS  = 0x%08lX  EFLAGS = 0x%08lX\n", rg.rbx, rg.ds, rg.eflags);
    logmsg("\tRCX = 0x%08lX  ES  = 0x%08lX\n", rg.rcx, rg.es);
    logmsg("\tRDX = 0x%08lX  FS  = 0x%08lX  FSB = 0x%08lX\n", rg.rdx, rg.fs, rg.fs_base);
    logmsg("\tRSI = 0x%08lX  GS  = 0x%08lX  GSB = 0x%08lX\n", rg.rsi, rg.gs, rg.gs_base);
    logmsg("\tRDI = 0x%08lX  RSP = 0x%08lX  RBP = 0x%08lX\n", rg.rdi, rg.rsp, rg.rbp);
    logmsg("\tR8  = 0x%08lX  R11 = 0x%08lX  OAX = 0x%08lX\n", rg.r8, rg.r11, rg.oax);
    logmsg("\tR9  = 0x%08lX  R12 = 0x%08lX  R14 = 0x%08lX\n", rg.r9, rg.r12, rg.r14);
    logmsg("\tR10 = 0x%08lX  R13 = 0x%08lX  R15 = 0x%08lX\n", rg.r10, rg.r13, rg.r15);
#else
    logmsg("\tEAX = 0x%08lX  ESI = 0x%08lX     EIP = 0x%08lX\n", rg.eax, rg.esi, rg.eip);
    logmsg("\tEBX = 0x%08lX  EDI = 0x%08lX  EFLAGS = 0x%08lX\n", rg.ebx, rg.edi, rg.eflags);
    logmsg("\tECX = 0x%08lX  EBP = 0x%08lX       `-> ....%c%c%c.%c%c.%c.%c.%c\n",
      rg.ecx,
      rg.ebp,
      FLAG_OVF(rg.eflags, 'o'),
      FLAG_DIR(rg.eflags, 'd'),
      FLAG_INTR(rg.eflags, 'i'),
      FLAG_SIGN(rg.eflags, 's'),
      FLAG_ZERO(rg.eflags, 'z'),
      FLAG_AUXC(rg.eflags, 'x'),
      FLAG_PAR(rg.eflags, 'p'),
      FLAG_CARRY(rg.eflags, 'c')
    );
    logmsg("\tEDX = 0x%08lX  ESP = 0x%08lX  OAX = 0x%08lX\n", rg.edx, rg.esp, rg.orig_eax);
    logmsg("\t CS = 0x%04hX   DS = 0x%04hX   SS = 0x%08hX\n", rg.cs, rg.ds, rg.ss);
    logmsg("\t_CS = 0x%04hX  _DS = 0x%04hX  _SS = 0x%08hX\n", rg.__cs, rg.__ds, rg.__ss);
    logmsg("\t ES = 0x%04hX   FS = 0x%04hX   GS = 0x%08hX\n", rg.es, rg.fs, rg.gs);
    logmsg("\t_ES = 0x%04hX  _FS = 0x%04hX  _GS = 0x%08hX\n", rg.__es, rg.__fs, rg.__gs);
#endif
    return;
}

static void dump_stack(pid_t pid) {
    struct user_regs_struct st;
    unsigned long ebp, ret;
    int i = 0;

    ptrace(PTRACE_GETREGS, pid, NULL, &st);
    logmsg("Stack backtrace dump =>\n");
#ifdef __x86_64__
    ebp = st.rbp;
#else
    ebp = st.ebp;
#endif

    do {
        ret = ptrace(PTRACE_PEEKDATA, pid, ebp + 4, NULL) ;
        logmsg("\t(#%d) EBP=0x%08lX, RET=0x%08lX\n", i++, ebp, ret);
        ebp = ptrace(PTRACE_PEEKDATA, pid, ebp, NULL);
    } while(ebp != 0 && ebp != (unsigned long)-1);
    return;
}

static int logmsg(char *fmt, ...) {
    va_list argp;
    int ret;

    va_start(argp, fmt);
    fprintf(conp, "\e[32mSEGV\e[1m>\e[0m "); // ]]]
    //fprintf(conp, "\e[32mSEGV|%d\e[1m>\e[0m ", getpid()); //]]]
    fprintf(logp, "SEGV> ");
    vfprintf(conp, fmt, argp);
    fflush(conp);
    ret = vfprintf(logp, fmt, argp);
    fflush(logp);

    va_end(argp);
    return ret;
}

static void *xlogger(void *arg) {
    char buf[MAXLNLEN];
    int *ptm = (int *)arg;
    fd_set rd;

    fcntl(ptm[0], F_SETFL, O_NONBLOCK);
    fcntl(ptm[1], F_SETFL, O_NONBLOCK);

    while(1) {
        int rv;
        FD_ZERO(&rd);
        FD_SET(ptm[0], &rd);
        FD_SET(ptm[1], &rd);

        /* We know that the fd in ptm[1] IS the highest of all values in
        ptm[]... it was opened last. */
        select(ptm[1] + 1, &rd, NULL, NULL, NULL);

        if(FD_ISSET(ptm[0], &rd)) {
            if((rv = read(ptm[0], buf, MAXLNLEN)) <= 0) { break; }
            fwrite(buf, rv, 1, stdout);
            fwrite(buf, rv, 1, logp);
        }

        if(FD_ISSET(ptm[1], &rd)) {
            if((rv = read(ptm[1], buf, MAXLNLEN)) <= 0) { break; }
            fwrite(buf, rv, 1, stdout);
            fwrite(buf, rv, 1, logp);
        }
    }

    return NULL;
}

/*-----------------------------------------------------------------------------
Process tree (#define PURE):
\_ bash
    \_ segvtracer
        \_ application
        \_ control thread
            \_ xlogger

Process tree (hook version):
\_ bash
    \_ application (#1: segvtracer)
        \_ application (#2: main)
        \_ control thread for segvtracer
            \_ xlogger

"Application #1" is the tracer, of which at fork() is made an exact copy of.
#2 sets "parent traces" mode (PTRACE_TRACEME) and then goes back to the
initialization code and to main() then.
xlogger is obviously created by the tracer part.

With GDmalloc, it becomes weird ;-)
\_ bash
    \_ segvtracer
        \_ application w/malloc tracer
        |   \_ worker
(       |   \_ control thread for application   )
(       |       \_ xxx                          )
        \_ control thread for segvtracer
            \_ xlogger

-------------------------------------------------------------------------------
File descriptors:
  tracer and xlogger:
    0 -> /dev/vc/1       (current tty)
    1 -> /dev/vc/1
    2 -> /dev/vc/1
    3 -> /dev/tty0       (FILE *conp, current too, but not redirectable since
                         created inside the program, descriptors 0-2 are passed
                         by the parent.)
    4 -> segvtracer.log  (FILE *logp, file copy of /dev/tty0)
    5 -> /dev/ptmx       (stdout link)
    6 -> /dev/ptmx       (stderr link)

  application:
    0 -> /dev/vc/1       (sharing with tracer's fd 0)
    1 -> /dev/pts/0      (stdout link)
    2 -> /dev/pts/1      (stderr link)
    3 -> /dev/tty0       (sharing with tracer's fd 3)
    4 -> segvtracer.log  (sharing with tracer's fd 4)

Note that ptmx is opened twice to allow for both stdout and stderr. This is
different than "the default file descriptors" 0 to 2, which all point to one
device.
*/

//==[ End of file ]============================================================
